ner_displacy_palette = {
    "CASENAME": "#da3650",
    "CITATION": "#67328b",
    "JUDGE": "#00a594",
    "COURT": "#fcd548",
    "PROVISION": "#007bac",
    "INSTRUMENT": "#6c63a5",
    "CONCEPT": "#df5a35",
}

ner_displacy_options = {
    "ents": [
        "CASENAME",
        "CITATION",
        "JUDGE",
        "COURT",
        "PROVISION",
        "INSTRUMENT",
        "CONCEPT",
    ],
    "colors": ner_displacy_palette,
}


dep_displacy_options = {
    "compact": False,
    "bg": "linear-gradient(90deg, #bdc3c7, #2c3e50)",
    "color": "white",
    "font": "Gotham",
}



import spacy
from spacy.pipeline import EntityRuler
from spacy.matcher import Matcher
from spacy.tokens import Doc
from collections import Counter
from blackstone.rules.concept_rules import CONCEPT_PATTERNS


class Concepts:
    """
    A custom pipeline component for identifying legal "concepts"
    in the doc. 

    This process is rules-based as opposed to statistical. 

    The doc is processed as usual, the EntityRuler is applied to the
    end of the pipeline, the patterns to look for are added to
    the ruler, matches are identified in the doc and are labelled 
    as CONCEPT entities. 

    The concept pipe is added after the EntityRuler and the concepts
    in the doc are accessible via doc._.concepts as a Counter object.
    
    Usage:
    concepts_pipe = Concepts(nlp)
    nlp.add_pipe(concepts_pipe)

    """

    def __init__(self, nlp) -> None:
        Doc.set_extension("concepts", default={}, force=True)
        self.ruler = EntityRuler(nlp)
        self.ruler.add_patterns(CONCEPT_PATTERNS)
        pipes = nlp.pipe_names
        if "EntityRuler" not in pipes:
            nlp.add_pipe(self.ruler, last=True)

    def __call__(self, doc: Doc) -> Doc:
        CONCEPTS = [ent.text.lower() for ent in doc.ents if ent.label_ == "CONCEPT"]
        if CONCEPTS:
            # CONCEPTS = list(set(CONCEPTS))
            CONCEPTS.sort()
            # Exclude terms with fewer than 5 occurences
            doc._.concepts = Counter(CONCEPTS).most_common(5)
        else:
            doc._.concepts = "None"
        return doc



"""
A slightly modified version of scispaCy's AbbreviationDetector. The only minor 
modification has been to handle common instances in legal text in which short form
abbreviations are bounded by quote marks, like so:

Section 1 of the Human Rights Act 1998 ("HRA") provides...

Save for this minor modification, the code has been untouched.

ScispaCy repo here -> https://github.com/allenai/scispacy
"""

from typing import Tuple, List, Optional, Set, Dict
from collections import defaultdict
from spacy.tokens import Span, Doc
from spacy.matcher import Matcher


def find_abbreviation(
    long_form_candidate: Span, short_form_candidate: Span
) -> Tuple[Span, Optional[Span]]:
    """
    Implements the abbreviation detection algorithm in "A simple algorithm
    for identifying abbreviation definitions in biomedical text.", (Schwartz & Hearst, 2003).

    The algorithm works by enumerating the characters in the short form of the abbreviation,
    checking that they can be matched against characters in a candidate text for the long form
    in order, as well as requiring that the first letter of the abbreviated form matches the
    _beginning_ letter of a word.

    Parameters
    ----------
    long_form_candidate: Span, required.
        The spaCy span for the long form candidate of the definition.
    short_form_candidate: Span, required.
        The spaCy span for the abbreviation candidate.

    Returns
    -------
    A Tuple[Span, Optional[Span]], representing the short form abbreviation and the
    span corresponding to the long form expansion, or None if a match is not found.
    """
    long_form = " ".join([x.text for x in long_form_candidate])
    short_form = " ".join([x.text for x in short_form_candidate])

    long_index = len(long_form) - 1
    short_index = len(short_form) - 1

    while short_index >= 0:
        current_char = short_form[short_index].lower()
        # We don't check non alpha-numeric characters.
        if not current_char.isalnum():
            short_index -= 1
            continue

            # Does the character match at this position? ...
        while (
            (long_index >= 0 and long_form[long_index].lower() != current_char)
            or
            # .... or if we are checking the first character of the abbreviation, we enforce
            # to be the _starting_ character of a span.
            (
                short_index == 0
                and long_index > 0
                and long_form[long_index - 1].isalnum()
            )
        ):
            long_index -= 1
            if long_index < 0:
                return short_form_candidate, None
        long_index -= 1
        short_index -= 1

    # If we complete the string, we end up with -1 here,
    # but really we want all of the text.
    long_index = max(long_index, 0)

    # Now we know the character index of the start of the character span,
    # here we just translate that to the first token beginning after that
    # value, so we can return a spaCy span instead.
    word_lengths = 0
    starting_index = None
    for i, word in enumerate(long_form_candidate):
        word_lengths += len(word)
        if word_lengths > long_index:
            starting_index = i
            break

    return short_form_candidate, long_form_candidate[starting_index:]


def containsQuotes(str, set):
    """ Check whether sequence str contains ANY of the items in set. """
    return 1 in [c in str for c in set]


def filter_matches(
    matcher_output: List[Tuple[int, int, int]], doc: Doc
) -> List[Tuple[Span, Span]]:
    # Filter into two cases:
    # 1. <Short Form> ( <Long Form> )
    # 2. <Long Form> (<Short Form>) [this case is most common].
    candidates = []
    for match in matcher_output:
        start = match[1]
        end = match[2]
        # Ignore spans with more than 8 words in.
        if end - start > 8:
            continue
        if end - start > 3:
            # Long form is inside the parens.
            # Take two words before.
            short_form_candidate = doc[start - 3 : start - 1]
            if short_form_filter(short_form_candidate):
                candidates.append((doc[start:end], short_form_candidate))
        else:
            # Normal case.
            # Short form is inside the parens.
            # Sum character lengths of contents of parens.
            QUOTES = ['"', "'", "‘", "’", "“", "”"]
            if containsQuotes(doc.text, QUOTES):
                abbreviation_length = sum([len(x) for x in doc[start:end]]) - 2
            else:
                abbreviation_length = sum([len(x) for x in doc[start:end]])
            max_words = min(abbreviation_length + 5, abbreviation_length * 2)
            # Look up to max_words backwards
            long_form_candidate = doc[max(start - max_words - 1, 0) : start - 1]
            candidates.append((long_form_candidate, doc[start:end]))
    return candidates


def short_form_filter(span: Span) -> bool:
    # All words are between length 2 and 10
    if not all([2 < len(x) < 10 for x in span]):
        return False
    # At least one word is alpha numeric
    if not any([x.is_alpha for x in span]):
        return False
    return True


class AbbreviationDetector:
    """
    Detects abbreviations using the algorithm in "A simple algorithm for identifying
    abbreviation definitions in biomedical text.", (Schwartz & Hearst, 2003).

    This class sets the `._.abbreviations` attribute on spaCy Doc.

    The abbreviations attribute is a `List[Span]` where each Span has the `Span._.long_form`
    attribute set to the long form definition of the abbreviation.

    Note that this class does not replace the spans, or merge them.
    """

    def __init__(self, nlp) -> None:
        Doc.set_extension("abbreviations", default=[], force=True)
        Span.set_extension("long_form", default=None, force=True)

        self.matcher = Matcher(nlp.vocab)
        self.matcher.add(
            "parenthesis", None, [{"ORTH": "("}, {"OP": "+"}, {"ORTH": ")"}]
        )
        self.global_matcher = Matcher(nlp.vocab)

    def find(self, span: Span, doc: Doc) -> Tuple[Span, Set[Span]]:
        """
        Functional version of calling the matcher for a single span.
        This method is helpful if you already have an abbreviation which
        you want to find a definition for.
        """
        dummy_matches = [(-1, int(span.start), int(span.end))]
        filtered = filter_matches(dummy_matches, doc)
        abbreviations = self.find_matches_for(filtered, doc)

        if not abbreviations:
            return span, set()
        else:
            return abbreviations[0]

    def __call__(self, doc: Doc) -> Doc:
        matches = self.matcher(doc)
        matches_no_brackets = [(x[0], x[1] + 1, x[2] - 1) for x in matches]
        filtered = filter_matches(matches_no_brackets, doc)
        occurences = self.find_matches_for(filtered, doc)

        for (long_form, short_forms) in occurences:
            for short in short_forms:
                short._.long_form = long_form
                doc._.abbreviations.append(short)
        return doc

    def find_matches_for(
        self, filtered: List[Tuple[Span, Span]], doc: Doc
    ) -> List[Tuple[Span, Set[Span]]]:
        rules = {}
        all_occurences: Dict[Span, Set[Span]] = defaultdict(set)
        already_seen_long: Set[str] = set()
        already_seen_short: Set[str] = set()
        for (long_candidate, short_candidate) in filtered:
            short, long = find_abbreviation(long_candidate, short_candidate)
            # We need the long and short form definitions to be unique, because we need
            # to store them so we can look them up later. This is a bit of a
            # pathalogical case also, as it would mean an abbreviation had been
            # defined twice in a document. There's not much we can do about this,
            # but at least the case which is discarded will be picked up below by
            # the global matcher. So it's likely that things will work out ok most of the time.
            new_long = long.string not in already_seen_long if long else False
            new_short = short.string not in already_seen_short
            if long is not None and new_long and new_short:
                already_seen_long.add(long.string)
                already_seen_short.add(short.string)
                all_occurences[long].add(short)
                rules[long.string] = long
                # Add a rule to a matcher to find exactly this substring.
                self.global_matcher.add(
                    long.string, None, [{"ORTH": x.text} for x in short]
                )
        to_remove = set()
        global_matches = self.global_matcher(doc)
        for match, start, end in global_matches:
            string_key = self.global_matcher.vocab.strings[
                match
            ]  # pylint: disable=no-member
            to_remove.add(string_key)
            all_occurences[rules[string_key]].add(doc[start:end])
        for key in to_remove:
            # Clean up the global matcher.
            self.global_matcher.remove(key)

        return [(k, v) for k, v in all_occurences.items()]



"""
Detects relationships between provisions and instruments identified by 
Blackstone's NER model. 

Example usage:

nlp = spacy.load("model")

TEXTS = [
    "Section 1 of the Theft Act 1968 sets out the definition of theft."
]
for text in TEXTS:
    doc = nlp(text) 
    relations = extract_legislation_relations(doc)
    for provision, provision_url, instrument, instrument_url in relations:
        print(f"\n{provision}\t{provision_url}\t{instrument}\t{instrument_url}")

"""
from typing import List, Tuple
import re
import requests
import time


def filter_spans(spans) -> List[str]:
    """
    Filter out overlapping spans. Returns a list of strings.
    """
    get_sort_key = lambda span: (span.end - span.start, span.start)
    sorted_spans = sorted(spans, key=get_sort_key, reverse=True)
    result = []
    seen_tokens = set()
    for span in sorted_spans:
        if span.start not in seen_tokens and span.end - 1 not in seen_tokens:
            result.append(span)
            seen_tokens.update(range(span.start, span.end))
    return result


def hasNumbers(inputString: str) -> bool:
    """
    Check if the provision candidate contains a digit.
    """
    return any(char.isdigit() for char in inputString)


def extract_legislation_relations(doc) -> List[Tuple]:
    """
    Extract relationships between provisions and instruments identified
    by Blackstone's NER with the assistance of the dependency parser. 

    This function receives a spaCy doc and returns a list of tuples, each 
    tuple containing the following elements:

    (provision, provision_URL, instrument, instrument_URL)
    """
    spans = list(doc.ents) + list(doc.noun_chunks)
    spans = filter_spans(spans)
    with doc.retokenize() as retokenizer:
        for span in spans:
            retokenizer.merge(span)

    relations = []
    for instrument in filter(lambda w: w.ent_type_ == "INSTRUMENT", doc):
        if instrument.dep_ in ("attr", "dobj"):
            subject = [w for w in instrument.head.lefts if w.dep_ == "nsubj"]
            if subject:
                subject = subject[0]
                if hasNumbers(subject):
                    # Get the URL for the instrument on legislation.gov.uk
                    target = set_legislation_target(instrument)
                    # Get the URL for the provision
                    if target:
                        provision = set_provision_target(target, subject)

                else:
                    provision = "None"
            relations.append((subject, provision, instrument, target))
        elif instrument.dep_ == "pobj" and instrument.head.dep_ == "prep":
            target = set_legislation_target(instrument)
            if target:
                if hasNumbers(str(instrument.head.head)):
                    provision = set_provision_target(target, instrument.head.head)
                    head = instrument.head.head
                else:
                    provision = "None"
                    head = "None"
            relations.append((head, provision, instrument, target))
    return relations


def set_legislation_target(instrument: str) -> str:
    """
    Returns the legislation.gov.uk for the identified instrument, 
    e.g. http://www.legislation.gov.uk/ukpga/1999/17/contents. 
    """
    if "Act" not in instrument.text:
        target_url = "None"
    elif "Act" in instrument.text:
        url = f"http://www.legislation.gov.uk/id?title={instrument.text}"
        page = requests.get(url)
        if page.status_code == 200:
            # allow time for the URL to resolve to the stable target
            time.sleep(0.1)
            target_url = page.url
        else:
            target_url = "None"

    url_targert = target_url

    return url_targert


def set_provision_target(url: str, subject: str) -> str:
    """
    Returns the legislation.gov.uk URL for the identified provision, 
    e.g. http://www.legislation.gov.uk/ukpga/1998/42/section/20.
    """
    url = url.replace("contents", "section/")
    provision_number = re.findall(r"\d+", str(subject))
    matches = [match for match in provision_number]
    if matches:
        url = url + str(matches[0])
    else:
        url = "None"
    return url



from setuptools import setup, find_packages


"""
Instructions for creating a release of the scispacy library.
1. Make sure your working directory is clean.
2. Make sure that you have changed the versions in "scispacy/version.py".
3. Create the distribution by running "python setup.py sdist" in the root of the repository.
4. Check you can install the new distribution in a clean environment.
5. Upload the distribution to pypi by running "twine upload <path to the distribution> -u <username> -p <password>".
   This step will ask you for a username and password - the username is "iclr-d" you can
   get the password from ICLR&D
"""

VERSION = {}
# version.py defines VERSION and VERSION_SHORT variables.
with open("blackstone/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

setup(
    name = 'blackstone',
    version = VERSION["VERSION"],
    url = 'https://research.iclr.co.uk/blackstone',
    author = 'ICLR&D (Incorporated Council of Law Reporting Research Lab)',
    author_email = 'research@iclr.co.uk',
    description = 'A SpaCy pipeline and models for NLP on unstructured legal text.',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords = ["law caselaw lawtech legaltech nlp spacy SpaCy biomedical"],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
         ],
    packages = find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    license="Apache",
    install_requires=[
        "spacy>=2.1.4",
        "requests",
        "conllu",
        "numpy",
        "pandas",
        ],
    tests_require=[
        "pytest",
        "pytest-cov",
        ],
    python_requires='>=3.6.0',
)
